import numpy as np

from gym_pybullet_drones.envs.BaseRLAviary import BaseRLAviary
from gym_pybullet_drones.utils.enums import DroneModel, Physics, ActionType, ObservationType

class HoverAviary(BaseRLAviary):
    """Single agent RL problem: hover at position."""

    ################################################################################
    
    def __init__(self,
                 drone_model: DroneModel=DroneModel.CF2X,
                 initial_xyzs=None,
                 initial_rpys=None,
                 physics: Physics=Physics.PYB,
                 pyb_freq: int = 240,
                 ctrl_freq: int = 30,
                 gui=False,
                 record=False,
                 obs: ObservationType=ObservationType.KIN,
                 act: ActionType=ActionType.RPM
                 ):
        """Initialization of a single agent RL environment.

        Using the generic single agent RL superclass.

        Parameters
        ----------
        drone_model : DroneModel, optional
            The desired drone type (detailed in an .urdf file in folder `assets`).
        initial_xyzs: ndarray | None, optional
            (NUM_DRONES, 3)-shaped array containing the initial XYZ position of the drones.
        initial_rpys: ndarray | None, optional
            (NUM_DRONES, 3)-shaped array containing the initial orientations of the drones (in radians).
        physics : Physics, optional
            The desired implementation of PyBullet physics/custom dynamics.
        pyb_freq : int, optional
            The frequency at which PyBullet steps (a multiple of ctrl_freq).
        ctrl_freq : int, optional
            The frequency at which the environment steps.
        gui : bool, optional
            Whether to use PyBullet's GUI.
        record : bool, optional
            Whether to save a video of the simulation.
        obs : ObservationType, optional
            The type of observation space (kinematic information or vision)
        act : ActionType, optional
            The type of action space (1 or 3D; RPMS, thurst and torques, or waypoint with PID control)

        """
        self.TARGET_POS = np.array([2,2,2])
        self.ARENA_SIZE = self.TARGET_POS*1.25
        self.EPISODE_LEN_SEC = 100


        super().__init__(drone_model=drone_model,
                         num_drones=1,
                         initial_xyzs=initial_xyzs,
                         initial_rpys=initial_rpys,
                         physics=physics,
                         pyb_freq=pyb_freq,
                         ctrl_freq=ctrl_freq,
                         gui=gui,
                         record=record,
                         obs=obs,
                         act=act
                         )

    ################################################################################
    
    def _computeReward(self):
        """Computes the current reward value.

        Returns
        -------
        float
            The reward.

        """
        state = self._getDroneStateVector(0)

        # ret = -10*(self.TARGET_POS[0]-state[0]) -10*(self.TARGET_POS[1]-state[1]) - 30*(self.TARGET_POS[2]-state[2])
        # ret = (self.TARGET_POS-state[0:3])
        # print("state x: ", state[0], "state y: ", state[1], "state z: ", state[2], "state 3: ", state[3])

        # time penalty
        norm_ep_time = (self.step_counter / self.PYB_FREQ) / self.EPISODE_LEN_SEC        
        time_penalty = -1*norm_ep_time
        ret = -10*(self.TARGET_POS[0]-state[0]) -10*(self.TARGET_POS[1]-state[1]) - 30*(self.TARGET_POS[2]-state[2]) + 20*time_penalty

        # if (abs(state[0]) > 0.8*self.ARENA_SIZE[0] or abs(state[1]) > 0.8*self.ARENA_SIZE[1] or abs(state[2]) > 0.8*self.ARENA_SIZE[2]):
        #     # boundary_penalty = 10 * np.linalg.norm(state[0:3] - self.ARENA_SIZE)
        #     ret += 100/np.linalg.norm(state[0:3] - self.ARENA_SIZE)
        
        if (abs(state[0]) > self.ARENA_SIZE[0] or abs(state[1]) > self.ARENA_SIZE[1] or abs(state[2]) > self.ARENA_SIZE[2]):
            # print ("--------------------------------------------------")
            # print ("   The drone has gone out of working boundary.    ")
            # print ("--------------------------------------------------")
            ret += -500
        
        # Truncate if the drone has tilted too much
        if (abs(state[7]) > .4 or abs(state[8]) > .4):
            # print ("--------------------------------------------------")
            # print ("           Getting tilting penalty!!!!            ")
            # print ("--------------------------------------------------")
            ret += -500
        
        # Truncate if the episode has timed out
        if self.step_counter / self.PYB_FREQ > self.EPISODE_LEN_SEC:
            # print ("--------------------------------------------------")
            # print ("   Time is up. Too long to reach to the goal.     ")
            # print ("--------------------------------------------------")
            ret += -30
        # if np.linalg.norm(self.TARGET_POS-state[0:3]) < 0.5:
        if np.linalg.norm(np.array([self.TARGET_POS[0]-state[0], self.TARGET_POS[1]-state[1], self.TARGET_POS[2]-state[2]])) < 0.25:
            ret += 50

        # ret = (max(0, 3 - np.linalg.norm(self.TARGET_POS-state[0:3]))*40)
        
        # ret = max(0, 2 - np.linalg.norm(self.TARGET_POS-state[0:3])**4)
        # print(ret)
        return ret

    ################################################################################
    
    def _computeTerminated(self):
        """Computes the current done value.

        Returns
        -------
        bool
            Whether the current episode is done.

        """
        state = self._getDroneStateVector(0)
        # if np.linalg.norm(self.TARGET_POS-state[0:3]) < 0.5:
        if np.linalg.norm(np.array([self.TARGET_POS[0]-state[0], self.TARGET_POS[1]-state[1], self.TARGET_POS[2]-state[2]])) < 0.25:
            print ("--------------------------------------------------")
            print ("         The drone has reached the goal.          ")
            print ("--------------------------------------------------")        
            return True
        else:
            return False
        
    ################################################################################
    
    def _computeTruncated(self):
        """Computes the current truncated value.

        Returns
        -------
        bool
            Whether the current episode timed out.

        """
        state = self._getDroneStateVector(0)
        # Truncate if the drone is outside the defined arena boundaries
        # if (abs(state[0]) > ARENA_SIZE_X or abs(state[1]) > ARENA_SIZE_Y or abs(state[2]) > ARENA_SIZE_Z
        # or abs(state[7]) > .4 or abs(state[8]) > .4):  # Truncate when the drone is too tilted
        if (abs(state[0]) > self.ARENA_SIZE[0] or abs(state[1]) > self.ARENA_SIZE[1] or abs(state[2]) > self.ARENA_SIZE[2]):
            print ("--------------------------------------------------")
            print ("   The drone has gone out of working boundary.    ")
            print ("--------------------------------------------------")

            return True
        
        # Truncate if the drone has tilted too much
        elif (abs(state[7]) > .4 or abs(state[8]) > .4):
            # print ("--------------------------------------------------")
            # print ("The drone has tilted too much. Resetting!!!!      ")
            # print ("--------------------------------------------------")
 
            return True
        
        # Truncate if the episode has timed out
        elif self.step_counter / self.PYB_FREQ > self.EPISODE_LEN_SEC:
            print ("--------------------------------------------------")
            print ("   Time is up. Too long to reach to the goal.     ")
            print ("--------------------------------------------------")

            return True
        else:
            return False

    ################################################################################
    
    def _computeInfo(self):
        """Computes the current info dict(s).

        Unused.

        Returns
        -------
        dict[str, int]
            Dummy value.

        """
        return {"answer": 42} #### Calculated by the Deep Thought supercomputer in 7.5M years
